package Basic;
//package basic

object basic {
  trait GCount[A] {
    def gcount : A => Int 
  }
  
  def genCount[A](x : A)(implicit count : GCount[A]) : Int = count.gcount(x)
  
  implicit object gcountInt extends GCount[Int] {
    def gcount = _ => 0
  }
  
  implicit object gcountUnit extends GCount[Unit] {
    def gcount = _ => 0
  }
  
  implicit def gcountPair[a,b](implicit countA : GCount[a], countB : GCount[b]) = new GCount[(a,b)] {
    def gcount = {
      case (x,y) => countA.gcount(x) + countB.gcount(y)
    }
  }
  
  implicit def gcountEither[a,b](implicit countA : GCount[a], countB : GCount[b]) = 
    new GCount[Either[a,b]] {
	  def gcount = {
      	case Left(x)   => countA.gcount(x)
        case Right(y)  => countB.gcount(y)
	  }
  	}
  
  implicit def gcountList[a](implicit countA : GCount[a]) : GCount[List[a]] = new GCount[List[a]] {
    def gcount = x => genCount(toList(x))
  }
  
  def toList[a] : List[a] => Either[Unit,(a,List[a])] = {
    case Nil      => Left ()
    case x :: xs  => Right((x,xs))
  }
  
  object sumInt extends GCount[Int] {
    def gcount = x => x
  }
  
  val x = genCount(List(1,2,3))(gcountList(sumInt))
  val y = genCount(List(1,2,3))
  
  // Support for higher-order kinded types
  
  
  sealed case class GRose[F[_],A](x : A, childs : F[GRose[F,A]])
  
  implicit object gcountList2 {
    def apply[a](implicit countA : GCount[a]) : GCount[List[a]] = gcountList[a](countA)
  }
  
  // LIMITATION: No type inference for higher-kinded types
  
  implicit def gcountGRose[F[_],A]
    (implicit countA : GCount[A], countF : {def apply[B] (implicit countB : GCount[B]) : GCount[F[B]]}) : GCount[GRose[F,A]] =
      new GCount[GRose[F,A]]{
        def gcount = {
          case GRose(x,xs) => countA.gcount(x) + countF(gcountGRose[F,A](countA,countF)).gcount(xs)
        }
      }
   
   def test(rose : GRose[List,Int])(implicit countR : (GCount[GRose[List,Int]])) : Int = countR.gcount(rose)

   def testCount : Int = test(myRose)(gcountGRose[List,Int](sumInt,gcountList2))   // there appears to be a bug here

   val myRose : GRose[List,Int] = GRose[List,Int](3,Nil)
  
  /*
  implicit def gcountRose[F[_],A](countA : GCount[A], countF : {def apply[B] (implicit countB : GCount[B]) : GCount[F[B]]})
          = new GCount[GRose[F,A]] {
    def gcount : GRose[F,A] => Int = {
      case GRose(x,xs) => countA.gcount(x) + countF(countA).gcount(xs)
     }
  }*/
  
  // First-class generic functions covered by the syb3 approach
  
  /*
  trait GMapT[A] {
    def gmapT : {def apply[b](x : b)(implicit di : GMapT[b]) : b} => A => A
  }
  
  implicit object gmapTInt extends GMapT[Int] {
    def gmapT = f => x => x
  }
  
  implicit def gmapTList[a](implicit countA : GMapT[a]) : GMapT[List[a]] = 
    */
}
