package Generic.GMTC;

import Generic._
import Isomorphism._
import scala._

object GMVisitor {
    trait Generic[G[_]] {
      def unit      : G[Unit]
      def int       : G[Int]
      def char      : G[Char]
      def plus[a,b] : G[a] => G[b] => G[Either[a,b]]
      def prod[a,b] : G[a] => G[b] => G[(a,b)]
      def view[a,b] : Iso[b,a] => (=> G[a]) => G[b]
    }
    
    trait Rep[T] {
      def accept[g[_]] (implicit gen : Generic[g]) : g[T]
    }
    
    implicit def RUnit = new Rep[Unit] {
      def accept[g[_]] (implicit gen : Generic[g]) =
        gen.unit
    }
    
    implicit def RInt = new Rep[Int] {
      def accept[g[_]] (implicit gen : Generic[g]) = 
        gen.int
    }
    
    implicit def RChar = new Rep[Char] {
      def accept[g[_]] (implicit gen : Generic[g]) = 
        gen.char
    }
    
    implicit def RPlus[a,b] (implicit a : Rep[a], b : Rep[b]) = 
      new Rep[Either[a,b]] {
        def accept[g[_]] (implicit gen : Generic[g]) = 
          gen.plus(a.accept[g](gen))(b.accept[g](gen))
      }

    implicit def RProd[a,b] (implicit a : Rep[a], b : Rep[b]) =
      new Rep[(a,b)] {
        def accept[g[_]] (implicit gen : Generic[g]) =
          gen.prod(a.accept[g](gen))(b.accept[g](gen))
      }
    
    //  A new generic function}
    case class Size[A](size : A => int)
	
    trait MySize extends Generic[Size] {
      def unit      = Size(x => 0)
      def int       = Size(x => 0)
      def char      = Size(x => 0)
      def plus[a,b] = a => b => Size(_.fold(a.size, b.size))
      def prod[a,b] = a => b => Size(x => a.size(x._1) + b.size(x._2))
      def view[a,b] = iso => a => Size(x => a.size(iso.from(x)))
    }

    implicit object mySize extends MySize
    
    def size[a] (x : a) (implicit rep : Rep[a]) = 
      rep.accept[Size].size(x)
    
    // inheritance of generic functions
    trait CountInt extends MySize {
      override def int = Size(x => 1)
    }
    
    def count[a] (x:a) (implicit rep : Rep[a]) =
      rep.accept[Size](new CountInt{}).size(x)
    
    //def test = Cons (3, Cons(4,Cons (5,Nil[int])))

    //def countTest = count (test)
}