package Generic.HODGP;

object MoorsHoDGP {
  def id[a] = (x:a) => x
  
  // Figure 1 of Moors paper (slightly adapted to conform to the latest version of Scala
  
  trait TC {type A; type B}
  
  trait BiFunctor[S <: BiFunctor[S]] extends TC {
    self : S => 
    def bimap[c,d](f : A => c, g : B => d) : S {type A=c; type B=d}
  }
  
  // case class Fix[S <: BiFunctor[S],a](out : S {type A=a; type B=Fix[S,a]}){

  trait Fix[S <: TC,a] {
    def map[b](f : a => b) : Fix[S,b]
    def cata[b](f : S {type A=a; type B=b} => b) : b
  }
  
  case class In[S <: BiFunctor[S],a](out : S {type A=a; type B=Fix[S,a]}) extends Fix[S,a] {
    def map[b](f : a => b) : Fix[S,b] = In(out.bimap(f,_.map(f)))
    def cata[b](f : S {type A=a; type B=b} => b) : b = f (out.bimap(id,_.cata[b](f)))
  }
  
  def ana[s <: BiFunctor[s],a,b](f : b => s {type A=a; type B=b})(x:b) : Fix[s,a] =
    In(f(x).bimap(id,ana[s,a,b](f)))
  
  def hylo[s <: BiFunctor[s],a,b,c](f : b => s {type A=a; type B=b}, g : s {type A=a; type B=c} => c)(x:b) : c =
    g(f(x).bimap(id,hylo[s,a,b,c](f,g)))
  
  trait Builder[S <: BiFunctor[S],a] {
    final def build() : Fix[S,a]=bf[Fix[S,a]](In[S,a])
    def bf[b](f : S {type A=a; type B=b} => b) : b
  }

  // Figure 3
  
  trait ListF extends BiFunctor[ListF]
  
  case class NilF[a,b] extends ListF {
    type A=a ; type B=b
    def bimap[c,d](f : a => c, g : b => d) : NilF[c,d] = NilF()
  }
  case class ConsF[a,b](x : a, xs : b) extends ListF {
    type A=a ; type B=b
    def bimap[c,d](f : a => c, g : b => d) : ConsF[c,d] =
      ConsF(f(x),g(xs))
  }
  
  type List[A] = Fix[ListF,A]
}
